"""
Voice connection utilities.

.. currentmodule:: curious.voice

.. autosummary::
    :toctree: voice

    voice_client
    voice_gateway
    voice_player
"""
from curious.voice.voice_client import VoiceClient
from curious.voice.voice_player import VoicePlayer
