"""
Classes that wrap objects returned by Discord.

A **dataclass** is a class that contains data and allows the user to edit classes. These classes are stateful, 
as in they store state which allows the user to edit them as appropriate.

.. currentmodule:: curious.dataclasses

.. autosummary::
    :toctree: dataclasses

    appinfo
    bases
    channel
    embed
    emoji
    guild
    invite
    member
    message
    permissions
    presence
    reaction
    role
    search
    user
    voice_state
    webhook
    widget

"""
